/*
 * This file is part of the AdvanceVGA project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "tsr.h"
#include "lib.h"

#include "card.h"

/* OEM information */
#define OEM_VENDOR_STR "Andrea Mazzoleni"
#define OEM_PRODUCT_STR "AdvanceVGA"
#define OEM_VERSION_NUM 0x0002
#define OEM_VERSION_STR "0.2 " __DATE__
#define OEM_STR OEM_VENDOR_STR " " OEM_PRODUCT_STR " " OEM_VERSION_STR

/***************************************************************************/
/* vga */

/* Free mode numbers */
#define VGA_CUSTOM_MODE 0x20

/* VGA frequencies */
#define VGA_DOTCLOCK_HIGH 28322000
#define VGA_DOTCLOCK_LOW 25175000

/* Freqeuencies pivot */
#define VGA_DOTCLOCK_PIVOT_0 (VGA_DOTCLOCK_LOW/4 + VGA_DOTCLOCK_HIGH/4)
#define VGA_DOTCLOCK_PIVOT_1 (VGA_DOTCLOCK_LOW/2 + VGA_DOTCLOCK_HIGH/4)
#define VGA_DOTCLOCK_PIVOT_2 (VGA_DOTCLOCK_LOW/2 + VGA_DOTCLOCK_HIGH/2)

void vga_mode_set(const CRTCPARAM* param) {
	STACK_DECL CRTCPARAM cp;
	unsigned pixel_per_clock;
	unsigned font_rows;
	int text_mode;

	memcpy(&cp,param,sizeof(CRTCPARAM));

	if (card_graph_get(0x06) & 0x1)
		text_mode = 0;
	else
		text_mode = 1;

	if (text_mode) {
		/* set the correct font for a 80 columns */
		if (cp.HDisp < 360)
			pixel_per_clock = 8;
		else if (cp.HDisp < 500)
			pixel_per_clock = 9;
		else if (cp.HDisp < 720)
			pixel_per_clock = 8;
		else
			pixel_per_clock = 9;
	} else
		pixel_per_clock = 4;

	cp.HDisp = cp.HDisp * 8 / pixel_per_clock;
	cp.HBStart = cp.HBStart * 8 / pixel_per_clock;
	cp.HSStart = cp.HSStart * 8 / pixel_per_clock;
	cp.HSEnd = cp.HSEnd * 8 / pixel_per_clock;
	cp.HBEnd = cp.HBEnd * 8 / pixel_per_clock;
	cp.HTotal = cp.HTotal * 8 / pixel_per_clock;
	if (!text_mode)
		cp.dotclockHz = cp.dotclockHz * 8 / pixel_per_clock;

	if (text_mode) {
		/* set the correct font for a 25 rows */
		regs_16 regs;
		font_rows = cp.VDisp / 25;
		if (font_rows < 13) {
			/* 8 rows */
			if (font_rows < 7)
				font_rows = 7;
			if (font_rows > 8)
				font_rows = 8;
			regs.x.ax = 0x1102;
			regs.x.bx = 0;
		} else if (font_rows < 15) {
			/* 14 rows */
			if (font_rows < 13)
				font_rows = 13;
			if (font_rows > 14)
				font_rows = 14;
			regs.x.ax = 0x1101;
			regs.x.bx = 0;
		} else {
			/* 16 rows */
			if (font_rows < 15)
				font_rows = 15;
			if (font_rows > 16)
				font_rows = 16;
			regs.x.ax = 0x1104;
			regs.x.bx = 0;
		}
		int_10_call(&regs);
	}

	card_signal_disable();
	card_crtc_all_set(&cp);
	card_doublescan_set(0);
	if (text_mode) {
		card_char_size_x_set(pixel_per_clock);
		card_char_size_y_set(font_rows);
	} else {
		card_char_size_x_set(8);
		card_char_size_y_set(cp.doublescan ? 2 : 1);
	}
	card_polarity_set(cp.hpolarity, cp.vpolarity);

	if (cp.dotclockHz <= VGA_DOTCLOCK_PIVOT_0) {
		card_divider_set(2);
		card_pll_set(0);
	} else if (cp.dotclockHz <= VGA_DOTCLOCK_PIVOT_1) {
		card_divider_set(2);
		card_pll_set(1);
	} else if (cp.dotclockHz <= VGA_DOTCLOCK_PIVOT_2) {
		card_divider_set(1);
		card_pll_set(0);
	} else {
		card_divider_set(1);
		card_pll_set(1);
	}

	card_signal_enable();
}

/***************************************************************************/
/* mode */

#define MODE_MAX 16

typedef struct mode_info_struct {
	CRTCPARAM param;
	unsigned effective_mode;
	unsigned reported_mode;
} mode_info;

static mode_info mode_map[MODE_MAX];

/* number of video mode defined */
static unsigned mode_max;

/* active video mode */
static mode_info* mode_active;

static unsigned mode_number_generator = VGA_CUSTOM_MODE;

unsigned mode_find_number(unsigned mode, unsigned x, unsigned y) {
	unsigned number;
	unsigned i;

	if (mode == 0x13 && x == 320 && y == 200)
		number = 0x13;
	else if (mode == 0x3)
		number = 0x3;
	else if (mode == 0x1)
		number = 0x1;
	else
		return mode_number_generator++;

	for(i=0;i<mode_max;++i)
		if (mode_map[i].reported_mode == number)
			break;
	if (i<mode_max)
		return mode_number_generator++;
	return number;
}

static char far* gettok_ptr = 0;

int _pascal gettok_separator(char c) {
	return c==' ' || c=='\t' || c=='=' || c=='(' || c==')';
}

char far* gettok(void) {
	char far* r;
	while (gettok_separator(*gettok_ptr))
		++gettok_ptr;
	if (!*gettok_ptr)
		return 0;
	r = gettok_ptr;
	while (*gettok_ptr && !gettok_separator(*gettok_ptr))
		++gettok_ptr;
	if (*gettok_ptr) {
		*gettok_ptr = 0;
		++gettok_ptr;
	}
	return r;
}

int mode_load(void) {
	STACK_DECL char buffer[128];

	int handle = open("vga.cfg",O_RDONLY);
	if (handle == -1) {
		cputs("Configuration file vga.cfg not found\r\n");
		return 0;
	}

	mode_max = 0;

	while (gets(handle, buffer, sizeof(buffer))) {
		char far* s;
		mode_info far* mode_ptr = mode_map + mode_max;

		/* too many modes */
		if (mode_max == MODE_MAX) {
			close(handle);
			cputs("Too many modelines\r\n");
			return 0;
		}

		gettok_ptr = buffer;

		memset(mode_ptr,0,sizeof(mode_info));

		s = gettok();
		if (!s) continue;

		if (memcmp(s,"modeline",8)!=0)
			continue;

		/* skip the name */
		s = gettok();
		if (!s) continue;

		s = gettok();
		if (!s) continue;
		if (strcmp(s,"vgaline")!=0)
			continue;

		s = gettok();
		if (!s) continue;
		mode_ptr->effective_mode = strtol(s,16);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.dotclockHz = strtold(s,1000000L);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HDisp = mode_ptr->param.width = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HSStart = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HSEnd = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HTotal = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VDisp = mode_ptr->param.height = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VSStart = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VSEnd = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VTotal = strtol(s,10);

		while ((s = gettok())!=0) {
			if (strcmp(s,"doublescan")==0)
				mode_ptr->param.doublescan = 1;
			if (strcmp(s,"interlaced")==0)
				mode_ptr->param.interlace = 1;
			if (strcmp(s,"-hsync")==0)
				mode_ptr->param.hpolarity = 1;
			if (strcmp(s,"-vsync")==0)
				mode_ptr->param.vpolarity = 1;
		}

		/* convert the format to the requested */
		if (mode_ptr->param.doublescan) {
			mode_ptr->param.VDisp *= 2;
			mode_ptr->param.VBStart *= 2;
			mode_ptr->param.VSStart *= 2;
			mode_ptr->param.VSEnd *= 2;
			mode_ptr->param.VBEnd *= 2;
			mode_ptr->param.VTotal *= 2;
		}
		if (mode_ptr->param.interlace) {
			mode_ptr->param.VDisp /= 2;
			mode_ptr->param.VBStart /= 2;
			mode_ptr->param.VSStart /= 2;
			mode_ptr->param.VSEnd /= 2;
			mode_ptr->param.VBEnd /= 2;
			mode_ptr->param.VTotal /= 2;
		}

		/* original */
		switch (mode_ptr->effective_mode) {
			case 0x13 :
				mode_ptr->param.bits = 8;
				mode_ptr->param.original_width = 320;
				mode_ptr->param.original_height = 200;
				break;
			case 0x3 :
				mode_ptr->param.bits = 16;
				mode_ptr->param.original_width = 80;
				mode_ptr->param.original_height = 25;
				break;
			case 0x1 :
				mode_ptr->param.bits = 16;
				mode_ptr->param.original_width = 40;
				mode_ptr->param.original_height = 25;
				break;
			default:
				close(handle);
				cputs("VGA mode ");
				cputu(mode_ptr->effective_mode,0,' ',16);
				cputs(" not supported\r\n");
				return 0;
		}

		mode_ptr->param.interlaceratio = 0;

		/* blank */
		mode_ptr->param.HBStart = mode_ptr->param.HDisp + 1;
		if (mode_ptr->param.HBStart > mode_ptr->param.HSStart)
			mode_ptr->param.HBStart = mode_ptr->param.HSStart;
		mode_ptr->param.HBEnd = mode_ptr->param.HSEnd + 1;
		mode_ptr->param.VBStart = mode_ptr->param.VDisp + 1;
		if (mode_ptr->param.VBStart > mode_ptr->param.VSStart)
			mode_ptr->param.VBStart = mode_ptr->param.VSStart;
		mode_ptr->param.VBEnd = mode_ptr->param.VSEnd + 1;

		/* number */
		mode_ptr->reported_mode = mode_find_number(mode_ptr->effective_mode,mode_ptr->param.width,mode_ptr->param.height);

		/* next */
		++mode_max;
	}

	close(handle);

	if (!mode_max) {
		cputs("No modelines found\r\n");
		return 0;
	}

	return 1;
}

void mode_print(void) {
	unsigned i;

	cputs("\r\nAvaliable modes:\r\n");
	for(i=0;i<mode_max;++i) {
		unsigned v;
		cputs("        ");
		cputu(mode_map[i].reported_mode,4,' ',16);
		cputs("h ");
		cputu(mode_map[i].param.width,4,' ',10);
		cputs("x");
		cputu(mode_map[i].param.height,4,' ',10);
		cputs(" ");

		v = mode_map[i].param.dotclockHz / (mode_map[i].param.HTotal * 100L);
		cputu(v / 10,2,' ',10);
		cputs(".");
		cputu(v % 10,1,' ',10);
		cputs(" kHz/");

		v = 10L * mode_map[i].param.dotclockHz / (mode_map[i].param.HTotal * (long)mode_map[i].param.VTotal);
		cputu(v / 10,2,' ',10);
		cputs(".");
		cputu(v % 10,1,' ',10);
		cputs(" Hz ");

		cputs("\r\n");
	}
}

/***************************************************************************/
/* init/done */

int vga_init(void) {

#ifndef NDEBUG
	cputs("SP: ");
	cputu(_SP,4,'0',16);
	cputs("\r\n");
	cputs("DS: ");
	cputu(_DS,4,'0',16);
	cputs("\r\n");
#endif

	if (!mode_load()) {
		return 0;
	}

#ifndef NDEBUG
	cputs("SP: ");
	cputu(_SP,4,'0',16);
	cputs("\r\n");
	cputs("DS: ");
	cputu(_DS,4,'0',16);
	cputs("\r\n");
#endif

	return 1;
}

void vga_done(void) {
}

/***************************************************************************/
/* vga_service */

unsigned vga_service_call_00;
unsigned vga_service_call_modex;

void vga_service_00(regs_16 far * regs) {
	unsigned req_mode;
	mode_info* mode_ptr;

	req_mode = regs->h.al;

	for(mode_ptr = mode_map;mode_ptr!=mode_map + mode_max;++mode_ptr)
		if (mode_ptr->reported_mode == req_mode)
			break;
	if (mode_ptr == mode_map + mode_max) {
		int_10_call(regs);
		return;
	}

	/* active mode */
	mode_active = mode_ptr;

	regs->h.al = mode_active->effective_mode;
	int_10_call(regs);
	regs->h.al = req_mode;

	vga_mode_set(&mode_active->param);

	++vga_service_call_00;
}
/*
--------V-M00400049--------------------------
MEM 0040h:0049h - VIDEO - CURRENT VIDEO MODE
Size:	BYTE
SeeAlso: MEM 0040h:004Ah,INT 10/AH=00h
*/
void vga_modex_check(void) {
	mode_info* mode_ptr;
	uint8 req_mode;
	unsigned hde;
	unsigned vde;
	uint8 d0;

	req_mode = *(uint8 far*)MK_FP(0x0040,0x004A);
	if (req_mode != 0x13) {
		/* only mode 13 for x mode */
		return;
	}

	hde = card_crtc_get(0x1);
	hde += 1;
	hde *= 4; /* default VGA pixel per clock */

	vde = card_crtc_get(0x12);
	vde |= (((unsigned)card_crtc_get(0x7)) << 7) & 0x100;
	vde |= (((unsigned)card_crtc_get(0x7)) << 3) & 0x200;
	vde += 1;

	d0 = card_crtc_get(0x09);
	/* adjust for doublescan */
	if (d0 & 0x80)
		vde /= 2;
	/* adjust for char y */
	vde /= (d0 & 0x1F) + 1;

	for(mode_ptr = mode_map;mode_ptr!=mode_map + mode_max;++mode_ptr)
		if (mode_ptr->effective_mode == req_mode && mode_ptr->param.width == hde && mode_ptr->param.height == vde)
			break;

	if (mode_ptr == mode_map + mode_max) {
		/* mode not found */
		return;
	}

	if (mode_active == mode_ptr) {
		/* already setup */
		return;
	}

	/* active mode */
	mode_active = mode_ptr;

	vga_mode_set(&mode_active->param);

	++vga_service_call_modex;
}

/***************************************************************************/
/* int */

void int_10_trap(regs_16 far * regs) {
	/* save all 32 bit registers, needed if the compiler use them */
	asm push eax
	asm push ebx
	asm push ecx
	asm push edx
	/* esi, edi aren't used by the compiler */

	if (regs->h.ah == 0x00) {
		vga_service_00(regs);
	} else
		int_10_call(regs);

	/* restore all 32 bit registers */
	asm pop edx
	asm pop ecx
	asm pop ebx
	asm pop eax
}

void int_21_hook(void) {
	/* save all 32 bit registers, needed if the compiler use them */
	asm push eax
	asm push ebx
	asm push ecx
	asm push edx
	/* esi, edi aren't used by the compiler */

	vga_modex_check();

	/* restore all 32 bit registers */
	asm pop edx
	asm pop ecx
	asm pop ebx
	asm pop eax
}

void int_8_hook(void) {
	/* save all 32 bit registers, needed if the compiler use them */
	asm push eax
	asm push ebx
	asm push ecx
	asm push edx
	/* esi, edi aren't used by the compiler */

	vga_modex_check();

	/* restore all 32 bit registers */
	asm pop edx
	asm pop ecx
	asm pop ebx
	asm pop eax
}

/***************************************************************************/
/* card */

int card_mode_callback(CRTCPARAM STACK_PTR *cp) {
	memcpy(cp,&mode_active->param,sizeof(CRTCPARAM));
	return 1;
}

/***************************************************************************/
/* TSR */

uint16 _pascal tsr_init_0(void) {
	return 1;
}

uint16 _pascal tsr_init_1(void) {
	int reset = 0;
	int i;

	cputs(OEM_PRODUCT_STR " by " OEM_VENDOR_STR " v" OEM_VERSION_STR "\r\n");

	if (!vga_init())
		return 0;

	if (!reset) {
		for(i=0;i<mode_max;++i) {
			if (mode_map[i].reported_mode == 0x3) {
				asm mov ax,0x3
				asm int 10h
				reset = 1;
				break;
			}
		}
	}

	if (!reset) {
		for(i=0;i<mode_max;++i) {
			if (mode_map[i].reported_mode == 0x1) {
				asm mov ax,0x1
				asm int 10h
				reset = 1;
				break;
			}
		}
	}

	if (reset)
		cputs(OEM_PRODUCT_STR " by " OEM_VENDOR_STR " v" OEM_VERSION_STR "\r\n");

	mode_print();

	return 1;
}

void _pascal tsr_done_0(void) {
}

void _pascal tsr_done_1(void) {
	vga_done();

	cputs(OEM_PRODUCT_STR " uninstalled");

	if (vga_service_call_00 || vga_service_call_modex) {
		cputs("\r\n\r\nVGA calls:\r\n");
		cputs("        ");
		cputu(vga_service_call_00,6,' ',10);
		cputs(" Mode Set\r\n");
		cputs("        ");
		cputu(vga_service_call_modex,6,' ',10);
		cputs(" ModeX Set\r\n");
	}
}